//==========================================================================================\\
//                                                                                          \\
//                                        EASY FOCUS                                        \\
//                                                                                          \\
//        Simple and clean search for the best focused slices in sequences of stacks        \\
//                                    (Macro for ImageJ)                                    \\
//                                                                                          \\
//                                                                                          \\
//     By using, copying, or altering source code in this file, you are agreeing to be      \\
//              bound by the following "Terms of Use" and "Legal Information".              \\
//                                                                                          \\
//============================= © 2019-2020 Bureau H. Gluender =============================\\


/*********************************************************************************************

	What For

		•	"Easy Focus" is an "ImageJ Macro".
		•	"Easy Focus" returns the best focused slices of stacks that represent focus
			series, no matter if a stack represents a single focus series (z-stack), a
			temporal sequence of focus series (4D-hyperstack), or focus series of colour
			channels (4D-hyperstack).
		•	"Easy Focus"-analyses can be restricted to a rectangular selection (ROI).
    	•	"Easy Focus"-analyses are RGB-agnostic and can't be applied to 5D-hyperstacks.


	Installation

		1.	Unzip the zip-Archive "EasyFocus_1-5-x.zip".
		2.	Move the plugin-file "Focus_LP.class" to the folder "ImageJ >> plugins".(*)
		3.	Move this macro-file "Easy_Focus.ijm" to the folder
			"ImageJ >> plugins >> macros".
			(Do not copy the file to the folder "ImageJ >> macros"!)
		4.	Quit and re-launch ImageJ.
		5.	Open an image-stack holding a single focus series, or a 4D-hyperstack holding
			a temporal sequence of focus series, or a 4D-hyperstack holding focus series
			of colour channels. The stack images or a selection (ROI) thereof need to be
			square-sized.
			(Please note that if your stacks do not conform to this size restriction, you
			may create conforming selections or regional analyses.)
		6.	ImageJ: Choose the item "Easy Focus" from the "Plugin"-submenu "macros".
			Fiji:   Choose the item "Easy Focus" from the "Plugin"-submenu "Macros".

	    (*)	  In case the plugin is missing, it can be downloaded from here:
			  <https://www.gluender.de/Miscellanea/MiscTexts/UtilitiesText.html#Gl-2015-1>


	Description

		1.	"Easy Focus" returns a 16bit image or image-stack that represents the slice(s)
			that is/are determined as being best focused with respect to the analysis of
			the whole image or of the indicated selection (ROI).
		2.	Its/their position(s) in the original stack is/are displayed in the subtitle
			bar(s).
		3.	Important note:
			The best focused slice needs not be perfectly in focus because it merely is
			the relatively best focused slice of a focus series according to a kind of
			spatial "majority decision" regarding the focus of the slice or selection.
			Consequently, the results will strongly depend on the depth-resolution of the
			focus series and on the image content.
		4.	For details of the applied approach see:
			<https://www.gluender.de/Miscellanea/MiscTexts/MiscExcerpts/ReadMe_FL.html>
		5.	The parameter setting for the analyses are fixed but it can be altered by
			holding down the "alt"-key when launching "Easy Focus".


	Terms of Use

		•	"Easy Focus" may be used without a license.
		•	"Easy Focus" is NOT in the public domain.
		•	"Easy Focus" may NOT be sold.
		•	"Easy Focus" is provided AS IS and you use it entirely at your own risk.
		•	"Easy Focus" calls a closed-source ImageJ-plugin that is contained within
			the zip-Archive "EasyFocus_1-5-x.zip".
			Its "Terms of Use" etc. can be accessed from here:
			<https://www.gluender.de/Miscellanea/MiscTexts/MiscExcerpts/ReadMe_FL.html>


	Legal Information

		•	"Easy Focus" is copyright-protected 2019/2020 by "Bureau H. Gluender".
		•	"Easy Focus" may be used without acquiring a license and it is definitely not
			allowed to sell the it.
		•	"Easy Focus" may not be (re-)distributed without the above "Terms of Use"
			and this "Legal Information".
		•	If source code in this file is altered and distributed, the above "Terms of
			Use" must be maintained and included with the altered code.
			It must also be mentioned that the altered code is based on copyrighted code
			of "Bureau H. Gluender" and that changes have been made.
		•	"Easy Focus" is given away AS IS. It is solely the responsibility of the
			user to determine its suitability for a particular purpose or use.
		•	Although "Easy Focus" has been developed with great care, you use it at
			your own risk.
		•	"Bureau H. Gluender" cannot be hold responsible for any kind of direct or
			indirect damage or loss——including but not limited to, any interruption of
			service, loss of business, loss of profits or good will, legal action, or any
			other consequential damages——that may be caused by installing, configuring or
			using "Easy Focus".
		•	Because the source code in this file can be arbitrarily modified at any time
			and to any extent, "Bureau H. Gluender" has no control over its state and use.
		•	"Easy Focus" is not a stand-alone software. It fully depends on ImageJ-soft-
			ware modules and their state of installation and development.
		•	By using "Easy Focus" the user agrees to the generally accepted fact that to
			date flawless software cannot be guaranteed. Hence, imperfect software is not
			faulty in the legal sense.
			"Easy Focus" may contain "bugs" that limit its use or make its use even
			impossible.

*********************************************************************************************/

//==========================================================================================\\
//==== IMAGEJ MACRO ========================================================================\\
//==========================================================================================\\

/********************************************************************************************/
macro "Easy Focus" {
/*********************************************************************************************

	Performs focus analyses of focus series that are provided as 3D-stacks or as 4D-hyper-
	stacks (temporal sequences of focus series, or focus series of color channels).
	The stack may show a rectangular selection.

		Focus analyses are performed by calling the plugin "Focus_LP".
	The parameter setting is fix but can be altered by holding down the "alt"-key during
	the launch of the macro, an action that presents the parameter dialog of "Focus_LP".
	(Changed parameter settings are not persistent.)

	A 16bit image or image-stack is returned that shows the slice(s) that is/are
	determined as being best focused. The original stack is left untouched.

		Requires a stack or 4D-hyperstack open in ImageJ.

	    Requires the ImageJ-PlugIn
	    	"Focus_LP.class"
		in the "plugins"-folder of the ImageJ installation.

	Version 0.1b0	(02. September 2019)
	Version 1.0.0	(05. September 2019)
	Version 1.0.1	styled option dialog message (14. September 2019)
	Version 1.0.2	rmvSuffix() (26. December 2019)
	Version 1.0.3	improved rmvSuffix() (21. January 2020)
	Version 1.0.4	channel specific processing (27. February 2020)
	Version 1.1b0	autoSelection() (29. February 2020)
	Version 1.1.0	Release (01. March 2020)
	Version 1.2b0	support need longer be a power of two (02. March 2020)
	Version 1.2.0	Release (02. March 2020)
	Version 1.5b0	no support restrictions (02. March 2020)
	Version 1.5.0	Release (xy. March 2020)

______________________________________________________________________________________________
                    Copyright-protected 2019-2020 by "Bureau H. Gluender"
*********************************************************************************************/
	requires( "1.52t" );
	macroTitle = "Easy Focus";
	cv = "1b5.0";

	check4Plugins( macroTitle, newArray( "Focus LP" ), newArray( "1.0.0" ) ); // v1.2.0
	if ( isKeyDown( "alt" ) ) std = false; else std = true;

	ch = false;
	msg1 = "An open 4D-hyperstack with more than three slices\n"
		 + "and at least one frame or channel is required.\n"
		 + "5D-hyperstacks cannot be processed.";
	if ( nImages < 1 ) exit( msg1 );
	Stack.getDimensions( w, h, c, s, f );
	if ( s < 3 || ( c > 1 && f > 1 ) ) exit( msg1 ); // v1.0.4
	if ( c > 1 && f < 2 ) { f = c; ch = true; } // v1.0.4
	if ( selectionType() < 0 ) sel = false; else sel = true; // v1.1.0
	hStck = getImageID();
	ttl = rmvSuffix( "" ); // v1.0.2

	if ( !std ) {
		helpLink = "https://www.gluender.de/Miscellanea/MiscTexts/UtilitiesHelp/Help_EF.html";
		criterion  = newArray( "Standard", "Diagonal", "Modified", "ModDiagonal" );
		kernelSize = newArray( "3", "5", "9", "17", "33", "65" );
		crtn = 0; // default
		kExp = 2; // default
		Dialog.create( macroTitle );
			Dialog.setInsets( -8, 229, 6 );
			Dialog.addMessage( "v" + cv, 9, "gray" );
			Dialog.setInsets( 0, 0, 8 );
			Dialog.addChoice( "Criterion", criterion, criterion[crtn] );
			Dialog.addChoice( "Size", kernelSize, kernelSize[kExp] );
			Dialog.addHelp( helpLink );
		Dialog.show();
		crtn = Dialog.getChoice();
		size = parseInt( Dialog.getChoice() );
	} else {
		crtn = "Standard";
		size = 9;
	}

	setBatchMode( true );
	for ( i=1; i<=f; i++ ) { // loop all frames or channels
		selectImage( hStck );
		run( "Select None" ); // v1.1.0
		if ( ch != true ) run( "Duplicate...", "duplicate frames=" + i );
			else run( "Duplicate...", "duplicate channels=" + i ); // v1.0.4
  		if ( bitDepth() != 16 ) run( "16-bit" );
  		if ( sel ) run( "Restore Selection" ); // v1.1.0
  		stck = getTitle();
   		run( "Focus LP", "criterion=" + crtn + " size=" + size );
   		returnStr = split( call( "Focus_LP.macroReturn" ), " " );
		run( "Select None" );
   		run( "Make Substack...", "slices=" + returnStr[0] );
   		close( stck );
  		rename( "img" );
  		resetMinAndMax();
   		setMetadata( "Label", "FocusSlice " + returnStr[0] );
   		if ( i > 1 ) run( "Concatenate...", "  title=r image1=r image2=img" ); else rename( "r" );
	}
	if ( ch ) { // v1.0.4
		run( "Make Composite", "display=Composite" );
		Stack.setDisplayMode( "grayscale" );
	}
	if ( sel ) run( "Restore Selection" ); // v1.1.0
	rename( ttl + "_result" );
	taw( cv, "\""+macroTitle+"\"", true );
	setBatchMode( false );
	exit;
}
//==========================================================================================\\
//==== IMAGEJ MACRO FUNCTIONS ==============================================================\\
//==========================================================================================\\
function check4Plugins( mNme, pNme, pVrs ) { // v1.2.0
	str = "ImageJ-PlugIn \"";
	List.setCommands;
	for ( i=0; i<pNme.length; i++ ) {
		if ( lengthOf( List.get( pNme[i] ) ) < 1 ) {
			exit( "Macro \"" + mNme + "\" requires " + str + pNme[i] + "\" !" );
		} else {
			a = split( call( replace( pNme[i], " ", "_" ) + ".macroReturn" ), " " );
			if ( vers2Int( a[a.length-1] ) < vers2Int( pVrs[i] ) ) {
				exit( "Please update " + str + pNme[i] + "\" to version " + pVrs[i] + " !" );
			}
		}
	}
}
function vers2Int( v ) { return parseInt( replace( replace( v, "b", "" ), "\\.", "" ) ); }
function rmvSuffix( nme ) { // v1.0.3
   if ( lengthOf( nme ) == 0 ) nme = getTitle;
   idx = lastIndexOf( nme, "." );
   if ( idx > -1 ) {
      sfx = nme.substring( idx );
      k = lengthOf( sfx );
      nme = substring( nme, 0, idx );
      if ( k < 6 && k > 2 && !matches( substring( sfx, 1, 2 ), "[0-9]" ) )
      	return nme;
      else { // v1.0.4
      	idx = lastIndexOf( sfx, "#" );
      	if ( idx > -1 ) return nme + "_" + substring( sfx, idx ); else return "BestFocused";
      }
   } else return nme;
}
function taw( v_4, v_3, dateTime ) {
	List.clear;
	if ( dateTime ) { getDateAndTime( yr, mth, weekday, day, h, mi, sc, ms );
	v_1  = "" + day + "d" + (mth+1) + "m" + yr + "y / " + h + "h" + mi + "m" + sc + "s"; }
	v_3 += fromCharCode(32,40,169,32,50,48,49,57,47,50,48,50,48,32,32,66,117,114,101,97,117);
	v_3 += fromCharCode(32,72,46,32,71,108,117,101,110,100,101,114,41);
	if (dateTime) List.set( fromCharCode(127,71,101,110,101,114,97,116,101,100,65,116),v_1 );
	List.set( fromCharCode(127,71,101,110,101,114,97,116,101,100,87,105,116,104),v_3 );
	List.set( fromCharCode(127,86,101,114,115,105,111,110),v_4 );
	setMetadata( "Info", List.getList() );
}
/*____________________________________________________________________________________________
                    Copyright-protected 2019-2020 by "Bureau H. Gluender"
*********************************************************************************************/